#include <errno.h>
#include <stdlib.h>// atoi
#include <stdio.h>// sscanf
#include <fcntl.h>// file flags
#include <string.h>// strncasecmp

#include "communication.h"

#ifdef WIN32
# include <winsock2.h>// closesocket, ioctlsocket
# include <ws2tcpip.h>
# pragma comment(lib, "wsock32.lib")
#else
#ifdef LINUX
# include <netinet/in.h>// sockaddr_in
# include <netdb.h>// gethostbyname
# include <unistd.h>// close
#include <sys/ioctl.h>// ioctl
# define SOCKET int
# define closesocket close
# define ioctlsocket ioctl
#endif// #ifdef LINUX
#endif// #ifdef WIN32

#define SOCKET_BUF_SIZE 65536// about 4 sec audio

static SOCKET sound_socket;

/** Initiate http response parameters structure */
void init_http_response_parameters()
{
    com_content_length = 0;
    com_content_type = NULL;
    com_shoutcast_metadata_period = 0;
}
/** Free http response parameters structure */
void free_http_response_parameters()
{
    if( com_content_type != NULL ) free( com_content_type );
    init_http_response_parameters();
}

/** Release resources and exit */
void disconnect()
{
    closesocket( sound_socket );
#ifdef WIN32
    WSACleanup();
#endif
}

/* Create socket and connect to server */
#if 1// using getaddrinfo
/** Return the last socket error */
static int socket_error()
{
#ifdef WIN32
    return WSAGetLastError();
#else
    return errno;
#endif
}
/** Create socket and connect it to specified server. @return 0 if success, -1 otherwise */
int connect_to_server()
{
    struct addrinfo *ai, *head, hints;
    char str_port[11];
    struct timeval tv;
    int res = -1;
#ifdef WIN32
    WSADATA wsaData;
#define WS_VERSION_CHOICE1 0x202// MAKEWORD(2, 2)
#define WS_VERSION_CHOICE2 0x101// MAKEWORD(1, 1)
    if( WSAStartup( WS_VERSION_CHOICE1, &wsaData ) != 0 &&
        WSAStartup( WS_VERSION_CHOICE2, &wsaData ) != 0 )
    {
        WEB_MSG("Could not initialize a socket\n");
        return -1;
    }

    if( wsaData.wVersion != WS_VERSION_CHOICE1 &&
        wsaData.wVersion != WS_VERSION_CHOICE2 )
    {
        WEB_MSG("Could not find WinSock1.1 and WinSock2.2\n");
        WSACleanup();
        return -1;
    }
#endif
    memset( &hints, 0, sizeof( hints ) );
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    snprintf( str_port, sizeof( str_port ), "%d", web_port );

    if( getaddrinfo( web_host, str_port, &hints, &head ) != 0 )
        return -1;

    for( ai = head; ai != NULL; ai = ai->ai_next )
    {
        if( (sound_socket = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol)) >= 0 )
        {
            if( connect(sound_socket, ai->ai_addr, ai->ai_addrlen) < 0 && 
                !(socket_error() == EINPROGRESS || socket_error() == EALREADY) )
            {
                closesocket( sound_socket );
                res = -1;
            }
            else
            {
                res = SOCKET_BUF_SIZE;
                setsockopt( sound_socket, SOL_SOCKET, SO_RCVBUF, (void *)&res, sizeof(res) );
                tv.tv_sec = 30;// 10 Secs Timeout
                tv.tv_usec = 0;
                setsockopt( sound_socket, SOL_SOCKET, SO_RCVTIMEO, (void*)&tv, sizeof(struct timeval) );
                res = 0;
                break;
            }
        }
    }
    if( head != NULL ) freeaddrinfo( head );
#ifdef WIN32
    if( res < 0 ) WSACleanup();
#endif
    return res;
}
#endif
#if 0// using gethostbyname
/** Create socket and connect it to specified server. @return 0 if success, -1 otherwise */
int connect_to_server()
{
    struct hostent *hp;
    struct sockaddr_in snd_srv_sa;
    struct timeval tv;
    int n;
#ifdef WIN32
    WSADATA wsaData;
#define WS_VERSION_CHOICE1 0x202// MAKEWORD(2, 2)
#define WS_VERSION_CHOICE2 0x101// MAKEWORD(1, 1)
    if( WSAStartup( WS_VERSION_CHOICE1, &wsaData ) != 0 &&
        WSAStartup( WS_VERSION_CHOICE2, &wsaData ) != 0 )
    {
        WEB_MSG("Could not initialize a socket\n");
        return -1;
    }

    if( wsaData.wVersion != WS_VERSION_CHOICE1 &&
        wsaData.wVersion != WS_VERSION_CHOICE2 )
    {
        WEB_MSG("Could not find WinSock1.1 and WinSock2.2\n");
        WSACleanup();
        return -1;
    }
#endif

    memset( &snd_srv_sa, 0, sizeof(struct sockaddr_in) );

    if( (hp = gethostbyname( web_host )) == NULL )
    {
        WEB_MSG("Unknown host: %s\n", web_host );
        return -1;
    }

    snd_srv_sa.sin_family = AF_INET;
    snd_srv_sa.sin_port   = htons( (unsigned short)web_port );

    memcpy( &snd_srv_sa.sin_addr, hp->h_addr_list[0], hp->h_length );
    /* Connect to the server */
    WEB_MSG("Connecting to server %s, port %d\n", web_host, web_port );
    /* Create socket */
    if( (sound_socket = socket( AF_INET, SOCK_STREAM, 0 )) < 0 )// PF_UNSPEC
    {
        perror("socket:");
        return -1;
    }
    if( connect( sound_socket, (struct sockaddr *)&snd_srv_sa, sizeof( struct sockaddr_in ) ) < 0 )
    {
        WEB_MSG("Could not connect to %s\n", web_host );
        perror("Connect");
        disconnect();
        return -1;
    }
    n = SOCKET_BUF_SIZE;
    setsockopt( sound_socket, SOL_SOCKET, SO_RCVBUF, (void *)&n, sizeof(n) );
    tv.tv_sec = 30;// 10 Secs Timeout
    tv.tv_usec = 0;
    setsockopt( sound_socket, SOL_SOCKET, SO_RCVTIMEO, (void*)&tv, sizeof(struct timeval) );
    return 0;
}
#endif
/** Send data to sound_socket. @return 0 if success, -1 if an error */
int send_data(const char* data, int len)
{
    int n;
    while( len > 0 )
    {
        if( (n = send(sound_socket, data, (size_t) len, 0)) < 0 )
        {
            WEB_MSG("Error while sending data.\n");
            return -1;
        }
        data += n;
        len -= n;
    }
    return 0;
}
/**
 * Read data from sound_socket to buf
 * @param buf - buffer to read data into
 * @param size - number of bytes to read
 * @return count of readed data
 */
size_t read_socket_data(char* buf, size_t size)
{
    //fd_set rd_set;
    struct timeval timeout;
    int n;
    size_t readed;
    //const int num_fds = sound_socket + 1;
    readed = 0;
    do
    {
        /* Waiting socket data */
        /* FIXME Waiting 5s maximum.
        timeout.tv_usec = 0;
        timeout.tv_sec = 5;// 5 sec timeout, because our buffer ~4 sec.
        FD_ZERO( &rd_set );
        FD_SET( sound_socket, &rd_set );
        n = select( num_fds, &rd_set, NULL, NULL, &timeout );
        // not > 0, because windows can return -1 for file sockets
        if( n == 0 )
        {
            WEB_MSG("Read timeout\n");
            return readed;
        }*/
        timeout.tv_usec = 1000;// + 1 ms
        timeout.tv_sec = 0;// wait another 1 ms, to get more data
        select( 0, NULL, NULL, NULL, &timeout );// usleep( 1000 );
        n = recv( sound_socket, buf, size - readed, 0 );
        if( n < 0 )
        {
            DPRINT("Error in read\n");
            perror("recv");
            return readed;
        }
        if( n == 0 )
        {
            DPRINT("End of file reached\n");
            return readed;
        }
        readed += (size_t) n;
        buf += n;
    }
    while( size > readed );
    return readed;
}
/** Read data from sound_socket to web_buf. @return size of the readed data or -1 if an error */
int read_socket()
{
    fd_set rd_set;
    struct timeval timeout;
    int n;
    /* Waiting socket data */
    const int num_fds = sound_socket + 1;
    FD_ZERO( &rd_set );
    FD_SET( sound_socket, &rd_set );
    timeout.tv_sec = 5;// 5 sec timeout, because our buffer ~4 sec.
    timeout.tv_usec = 0;
    n = select( num_fds, &rd_set, NULL, NULL, &timeout );
    // not > 0, because windows can return -1 for file sockets
    if( n != 0 )
    {
        timeout.tv_sec = 0;// wait another 1 ms, to get more data
        timeout.tv_usec = 1000;// + 1 ms
        select( 0, NULL, NULL, NULL, &timeout );// usleep( 1000 );
        n = recv( sound_socket, &web_buf[0], BUF_SIZE, 0 );
        if( n < 0 )
        {
            DPRINT("Error in read\n");
            perror("recv");
            return -1;
        }
        if( n == 0 )
        {
            DPRINT("End of file reached\n");
            return -1;
        }
        return n;
    }
    // n = 0, timeout
    WEB_MSG("Read timeout\n");
    return -1;
    
    /* for debugging
    if( (n = read( fd, &web_buf[pos], BUF_SIZE - pos )) <= 0 )
    {
        DPRINT("Error in read.\n");
        return -1;
    }
    */
    return n;
}
/** Read line from sound_socket to web_buf. @param n - start position */
int read_socket_line(size_t n)
{
    char ch;
    int res;
    for( ; n < MAX_HTTP_LINE_LENGTH; n++ )
    {
        res = recv( sound_socket, &ch, 1, 0 );
        if( res < 0 )
        {
            DPRINT("Error in read\n");
            perror("recv");
            return -1;
        }
        if( res == 0 )
        {
            DPRINT("End of file reached\n");
            return -1;
        }
        if( ch == '\n' )
        {
            web_buf[n] = '\0';
            if( n > 0 && web_buf[n - 1] == '\r' )
                web_buf[n - 1] = '\0';

            DPRINT( "%s\n", &web_buf[0] );
            return (int)++n;
        }
        web_buf[n] = ch;
    }
    WEB_MSG("HTTP line too long!\n");
    return -1;
}
/** @return -1 - server error, 0 - 200 OK*/
static int http_check_status()
{
    int res;
    if
    (
        sscanf( &web_buf[0], "HTTP/%*s%i", &res ) == 1 ||
        sscanf( &web_buf[0], "RTSP/%*s%i", &res ) == 1 ||
        sscanf( &web_buf[0], "ICY %i", &res ) == 1
    )
    {
        if( res == 200 )
            return 0;
        WEB_MSG( "Server status: %s\n", &web_buf[0] );
        return -1;
    }
    WEB_MSG( "Unknown status: %s\n", &web_buf[0] );
    return -1;
}
/** @param pos - start position in the web_buf, @return 0 if success, -1 if an error */
int read_response_in_pos( size_t pos, int(*line_parser)() )
{
    int len;

    len = read_socket_line( pos );
    if( len < 0 )
        return -1;
    if( http_check_status() < 0 )
        return -1;
    /* more lines */
    while( 1 )
    {
        len = read_socket_line( 0 );
        if( len < 0 )
            return -1;
        if( len <= 2 )
            return 0;// '\r\n\r\n' was found
        if( line_parser() < 0 )
            return -1;
    }
}
/** Read server response. @return 0 if success, -1 if an error */
int read_response( int(*line_parser)() )
{
    return read_response_in_pos( 0, line_parser );
}
/** Skip content. @return 0 if success, -1 if an error */
int skip_content()
{
    int n;
    size_t len;

    len = 0;
    while( len < com_content_length )
    {
        if( (n = read_socket()) < 0 )
        {
            DPRINT("Skip content reading error\n");
            return -1;
        }
        len += (size_t)n;
    }
    return 0;
}
/** Set socket buffer size and fill it with audio data */
int prebuffering()
{
    size_t num;
    struct timeval timeout;
    do
    {
        timeout.tv_sec = 0;
        timeout.tv_usec = 500000;
        select( 0, NULL, NULL, NULL, &timeout );
        if( ioctlsocket( sound_socket, FIONREAD, &num ) < 0 )
        {
            WEB_MSG("Error while getting size of the buffered data.\n");
            return -1;
        }
        WEB_MSG( "%u%%\r", (100 * num) / (SOCKET_BUF_SIZE - SOCKET_BUF_SIZE / 8) );
    } while ( num < (SOCKET_BUF_SIZE - SOCKET_BUF_SIZE / 8) );
    return 0;
}
/** Out web_buf data to stdout. @return 0 if success, -1 if an error */
int out_sound_stream(size_t len)
{
    int n;
    const char* data = &web_buf[0];
    do
    {
        if( (n = write( STDOUT_FILENO, data, len )) < 0 )
        {
            WEB_MSG("Error while sending sound data.\n");
            return -1;
        }
        data += n;
        len -= (size_t) n;
    }
    while( len > 0 );
    return 0;
}
