#ifndef _WEB_RADIO_COMMUNICATION_H_
#define _WEB_RADIO_COMMUNICATION_H_

#include "common.h"// definitions

/** Buffer size to receive server responses */
#define BUF_SIZE 8192
/** Max HTTP line length. Line length may contents base64 encoded sound header data */
#define MAX_HTTP_LINE_LENGTH (BUF_SIZE)// including trailing '\n'

/** Buffer to receive a server responses and to send request */
char web_buf[BUF_SIZE];
/* Common HTTP response parameters */
/** Content length */
size_t com_content_length;
/** Content type */
char* com_content_type;
/** SHOUTcast/Icecast metaint parameter */
size_t com_shoutcast_metadata_period;
//size_t com_audio_bit_rate;// to calculate prebuffer size

/** Initiate http response parameters structure */
void init_http_response_parameters();
/** Free http response parameters structure */
void free_http_response_parameters();
/** Create socket and connect it to specified server. @return 0 if success, -1 otherwise */
int connect_to_server();
/** Release resources and exit */
void disconnect();
/**
 * Read data from sound_socket to buf
 * @param buf - buffer to read data into
 * @param size - number of bytes to read
 * @return count of readed data
 */
size_t read_socket_data(char* buf, size_t size);
/** Read data from sound_socket to web_buf. @return size of the readed data or -1 if an error */
int read_socket();
/** Read line from sound_socket to web_buf. @param n - start position */
int read_socket_line(size_t n);
/** @param pos - start position in the web_buf, @return 0 if success, -1 if an error */
int read_response_in_pos( size_t pos, int(*line_parser)() );
/** Read server response. @return 0 if success, -1 if an error */
int read_response( int(*line_parser)() );
/** Skip content. @return 0 if success, -1 if an error */
int skip_content();
/** Set socket buffer size and fill it with audio data */
int prebuffering();
/** Send data to sound_socket. @return 0 if success, -1 if an error */
int send_data(const char* data, int len);
/** Out web_buf data to stdout. @return 0 if success, -1 if an error */
int out_sound_stream(size_t len);

#endif//_WEB_RADIO_COMMUNICATION_H_
