/*
HTTP server response dispatcher
 */
#include <string.h>// strlen, strdup, strnicmp
#include <stdlib.h>// atoi

#include "http.h"
#include "common.h"
#include "communication.h"
#include "icecast-shoutcast.h"
#include "wma.h"

#define HTTP_DEFAULT_PORT 80

/*
SHOUcast server response example:
ICY 200 OK
icy-notice1:<BR>This stream requires <a href="http://www.winamp.com/">Winamp</a><BR>
icy-notice2:SHOUTcast Distributed Network Audio Server/Linux v1.9.8<BR>
icy-name:rockradio.de
icy-genre:Rock
icy-url:http://www.rockradio.de
content-type:audio/mpeg
icy-pub:1
icy-metaint:8192
icy-br:128
*/

/** @return -1 - error, 0 - success */
static int http_header_line_parser()
{
    int tmp;
    size_t pos = 0;
    if( strnicmp(&web_buf[0], "Content-Length:", strlen("Content-Length:")) == 0 )
    {
        pos += strlen("Content-Length:");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        tmp = atoi( &web_buf[pos] );
        if( tmp < 0 )
        {
            WEB_MSG("Bad header: \"%s\"\n", &web_buf[pos]);
            return -1;
        }
        else
            com_content_length = (size_t) tmp;
    } else if( strnicmp(&web_buf[pos], "Content-Type:", strlen("Content-Type:")) == 0 )
    {
        pos += strlen("Content-Type:");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        com_content_type = strdup( &web_buf[pos] );
        if( com_content_type == NULL )
            return -1;
        DDPRINT("Content-Type is \"%s\"\n", com_content_type);
    } else if( strnicmp(&web_buf[pos], "icy-metaint:", strlen("icy-metaint:")) == 0 )
    {
        pos += strlen("icy-metaint:");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        tmp = atoi( &web_buf[pos] );
        if( tmp <= 0 )
        {
            WEB_MSG("Bad header: \"%s\"\n", &web_buf[pos]);
            return -1;
        }
        com_shoutcast_metadata_period = (size_t) tmp;
        DDPRINT("Shoutcast metadata period: %u\n", com_shoutcast_metadata_period);
    }
    return 0;
}

int http()
{
    int n;
    /* Check the port number */
    if( web_port <= 0 || web_port > 65535 ) web_port = HTTP_DEFAULT_PORT;
    /* Connecting to the server */
    if( connect_to_server() < 0 ) return -1;
    /* Sending GET request depends from connection type */
    DPRINT("Sending GET request\n");

#define REQUEST_PROXY_STR "GET http://%s:%d/%s HTTP/1.0\r\n" \
            "HOST: %s\r\n" \
            "User-Agent: "APP_NAME"\r\n" \
            "Accept: */*\r\n" \
            "Icy-MetaData: 1\r\n" \
            "Connection: keep-alive\r\n" \
            "\r\n"

#define REQUEST_STR "GET /%s HTTP/1.0\r\n" \
            "Host: %s\r\n" \
            "Accept: */*\r\n" \
            "User-Agent: "APP_NAME"\r\n" \
            "Icy-MetaData: 1\r\n" \
            "Connection: keep-alive\r\n" \
            "\r\n"

    /*if( proxy )
        n = snprintf( web_buf, BUF_SIZE - 1, REQUEST_PROXY_STR,
                                      web_host, web_port, web_file, web_host );
    else*/
        n = snprintf( web_buf, BUF_SIZE - 1, REQUEST_STR, web_file, web_host );
    DPRINT( &web_buf[0] );
    if( (n = send_data(web_buf, n)) >= 0 )
    {
        /* Reading response and choose protocol handler */
        DPRINT("Reading response\n");
        init_http_response_parameters();
        n = read_response( http_header_line_parser );
        if( n >= 0 )
        {
            if( com_content_type != NULL )
            {
                if( strnicmp(com_content_type, "audio/", strlen("audio/")) == 0 )
                {
                    n = icecast_shoutcast();
                } else if
                    ( strnicmp(com_content_type,
                        "application/vnd.ms.wms-hdr.asfv1",
                        strlen("application/vnd.ms.wms-hdr.asfv1")) == 0 ||
                      strnicmp(com_content_type,
                        "application/x-mms-framed",
                        strlen("application/x-mms-framed")) == 0 ||
                      strnicmp(com_content_type,
                        "application/octet-stream",
                        strlen("application/octet-stream")) == 0 ||
                      strnicmp(com_content_type,
                        "video/x-ms-asf",
                        strlen("video/x-ms-asf")) == 0 ||
                      strnicmp(com_content_type,
                        "video/x-ms-wmv",
                        strlen("video/x-ms-wmv")) == 0
                    )
                {
                    DPRINT("Windows Media Server protocol is detected\n");
                    n = wma();
                } else
                {
                    WEB_MSG("Unknown protocol\n");
                    n = -1;
                }
            } else
            {
                WEB_MSG("Content-type is unknown\n");
                n = -1;
            }
        }
        free_http_response_parameters();
    }
    disconnect();
    WEB_MSG("Good Bye!\n");
    return n;
}
