/*
Simple Icecast/Shoutcast client.
*/
#include "icecast-shoutcast.h"
#include "common.h"
#include "communication.h"

int icecast_shoutcast()
{
    size_t n;
    size_t block;
    size_t read_pos;
    DPRINT("SHOUTcast/Icecast protocol is detected\n");
    DPRINT("Start prebuffering\n");
    /* Prebuffering */
    if( prebuffering() < 0 )
        return -1;
    WEB_MSG("Prebuffering complete, start playing\n");
    /* If metadata period == 0, just send sound data */
    if( com_shoutcast_metadata_period == 0 )
    {
        DPRINT("Reading from socket\n");
        while( !web_quit_request )
        {
            n = read_socket_data( &web_buf[0], BUF_SIZE );
            if( n < BUF_SIZE )
                return -1;

            DDPRINT( "Readed: %u\n", n );

            if( out_sound_stream( n ) < 0 )
                return -1;
        }
        return 0;
    }
    /* If metadata period > 0, read sound block and metadata */
    while( !web_quit_request )
    {
        /* Send sound data */
        read_pos = 0;
        do
        {
            block = com_shoutcast_metadata_period - read_pos;
            if( block > BUF_SIZE )
                block = BUF_SIZE;

            n = read_socket_data( &web_buf[0], block );

            DDPRINT( "Read position: %u, Readed: %u...", read_pos, n );

            if( out_sound_stream( n ) < 0 )
                return -1;

            DDPRINT("Sended.\n");

            read_pos += n;
            if( n != block )
                return -1;
        }
        while( read_pos < com_shoutcast_metadata_period );
        /* Get metadata length, max size is 255 * 16 = 4080 bytes without zero */
        /* To save memory also can use web_buf */
        if( read_socket_data( &web_metadata_buf[0], 1 ) != 1 )
            return -1;
        block = ((size_t)web_metadata_buf[0]) << 4;// size of the metadata
        if( block > 0 )
        {
            /* Read metadata */
            n = read_socket_data( &web_metadata_buf[0], block );
            if( n != block )
                return -1;
            web_metadata_buf[block] = '\0';// end of line
            /* Print metadata */
            /* StreamTitle='title';StreamUrl='url';...*/
            WEB_MSG( "%s\n", &web_metadata_buf[0] );
        }
    }
    return 0;
}
