/*
Simple Windows media audio HTTP streaming protocol client.
*/
#include <string.h>// strdup, strlen, strncasecmp
#include <stdlib.h>// atoi, free

#include "common.h"
#include "wma.h"
#include "communication.h"

#define MMS_HEADER_SIZE 8

static char* web_features;

/** Initiate http response parameters structure */
static void init_wma_response_parameters()
{
    init_http_response_parameters();
    web_features = NULL;
}
/** Free http response parameters structure */
static void free_wma_response_parameters()
{
    if( web_features != NULL ) free( web_features );
    init_wma_response_parameters();
}

/** @return -1 - error, 0 - success */
static int wma_header_line_parser()
{
    int tmp;
    char* p;
    size_t pos = 0;
    if( strnicmp(&web_buf[pos], "Content-Length:", strlen("Content-Length:")) == 0 )
    {
        pos += strlen("Content-Length:");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        tmp = atoi( &web_buf[pos] );
        if( tmp < 0 )
            DPRINT("Bad header: \"%s\"\n", &web_buf[pos]);
        else
            com_content_length = (size_t) tmp;
    } else if( strnicmp(&web_buf[pos], "Content-Type:", strlen("Content-Type:")) == 0 )
    {
        pos += strlen("Content-Type:");
        while( web_buf[pos] == ' ' ) pos++;// skip spaces
        com_content_type = strdup( &web_buf[pos] );
        if( com_content_type == NULL )
            return -1;
        DDPRINT("Content-Type is \"%s\"\n", com_content_type);
    } else if( strnicmp(&web_buf[pos], "Pragma:", strlen("Pragma:")) == 0 )
    {
        pos += strlen("Pragma:");
        // FIXME this field is separated with a comma ','
        // but what happened, if we will have: Pragma:param="...,...", ?
        do
        {
            while( web_buf[pos] == ' ' ) pos++;// skip spaces
            if( strnicmp( &web_buf[pos], "features=", strlen("features=")) == 0 )
            {
                pos += strlen("features=");
                //p = &web_buf[pos];
                if( web_buf[pos] == '"' )
                    p = strstr( &web_buf[++pos], "\"" );
                else
                    p = strstr( &web_buf[pos], "," );
                if( p != NULL )
                    *p = '\0';
                web_features = strdup( &web_buf[pos] );
                if( web_features == NULL )
                    return -1;
                DDPRINT( "Pragma features: %s\n", web_features );
                break;
            }
            while( web_buf[pos] != ',' && web_buf[pos] != '\0' ) pos++;
        }
        while( web_buf[pos++] == ',' );
    }
    return 0;
}

static int play()
{
    size_t size;// uint16
    size_t n;
    size_t block;
    size_t read_pos;
    int reason;// uint32

    /* Prebuffering */
    DPRINT("Start prebuffering\n");
    if( prebuffering() < 0 ) return -1;
    WEB_MSG("Prebuffering complete, start playing\n");
    while( !web_quit_request )
    {
        n = read_socket_data( &web_buf[0], 4 );// 4 = size position in the header
        if( n < 4 )
            return -1;
        /* Decode framing header */
        if( web_buf[0] != 0x24 )// or '$'
        {
            WEB_MSG("Chunk not recognized: got 0x%02x\n%s", web_buf[0], &web_buf[0] );
            return -1;
        }
        //packet_id = web_buf[1];// type of the packet
        size = (unsigned char)web_buf[2] | ((unsigned char)web_buf[3] << 8);// uint16
        DDPRINT( "Got packet type: %c\n", web_buf[1] );
        switch( web_buf[1] )
        {
        case 'H'/*0x48*/:// $H (Header) Packet
            if( size < MMS_HEADER_SIZE )
            {
                DPRINT( "Incorrect $H chunk size: %u\n", size );
                return -1;
            }
            size -= MMS_HEADER_SIZE;// mms header size
            /* Decode MMS header */
            n = read_socket_data( &web_buf[0], MMS_HEADER_SIZE );
            if( n < MMS_HEADER_SIZE )
                return -1;
            // uint32 location_id = web_buf[0]|(web_buf[1]<<8)|(web_buf[2]<<16)|(web_buf[3]<<24);
            // 37 a6 1c 46, 37 a6 1d c6
            // incarnation = web_buf[4];// ignore
            // ca
            // uint8 afflags = web_buf[5];// usially 0xfe
            // fe
            // packet_size = (unsigned char)web_buf[6]|((unsigned char)web_buf[7] <<8 );// uint16
            // d3 ad
            // payload is asf header
            break;
        case 'D'/*0x44*/:// $D (Data) Packet
            if( size < MMS_HEADER_SIZE )
            {
                DPRINT( "Incorrect $D chunk size: %u\n", size );
                return -1;
            }
            size -= MMS_HEADER_SIZE;
            /* Decode MMS header */
            n = read_socket_data( &web_buf[0], MMS_HEADER_SIZE );
            if( n < MMS_HEADER_SIZE )
                return -1;
            // uint32 location_id = web_buf[0]|(web_buf[1]<<8)|(web_buf[2]<<16)|(web_buf[3]<<24);
            // 37 a6 1c 46, 37 a6 1d c6
            // incarnation = web_buf[4];// ignore
            // ca
            // uint8 afflags = web_buf[5];// usially 0xfe
            // fe
            // packet_size = web_buf[6] | (web_buf[7] << 8);// uint16
            // d3 ad
            // payload is sound data
            break;
        case 'C'/*0x43*/:// $C (Stream Change Notification) Packet
            if( size != 4 )// reason must be 0x400D14BE or 0
            {
                DPRINT( "Incorrect $C chunk size: %u\n", size );
                return -1;
            }
            continue;
            break;
        case 'E'/*0x45*/:// $E (End-of-Stream Notification) Packet
            if( size != 4 )
            {
                DPRINT( "Incorrect $E packet size: %u\n", size );
                return -1;
            }
            n = read_socket_data( &web_buf[0], 4 );
            if( n < 4 )
                return -1;
            reason =  (unsigned char)web_buf[0] |
                     ((unsigned char)web_buf[1] << 8) |
                     ((unsigned char)web_buf[2] << 16) |
                     ((unsigned char)web_buf[3] << 24);
            WEB_MSG("End of stream packet was received, exiting\nReason: ");
            switch( reason )
            {
            case 0x00000000:
                WEB_MSG("S_OK\n");
                break;
            case 0x00000001:
                WEB_MSG("S_FALSE\n");
                break;
            case 0x000D2F09:
                WEB_MSG("NS_S_EOS_RECEDING\n");
                break;
            default:
                WEB_MSG( "0x%08x\n", reason );
                break;
            }
            return 0;
            break;
        case 'M'/*0x4d*/:// $M (Metadata) Packet
            if( size < MMS_HEADER_SIZE )
            {
                DPRINT( "Incorrect $M chunk size: %u\n", size );
                return -1;
            }
            size -= MMS_HEADER_SIZE;
            /* Decode MMS header */
            n = read_socket_data( &web_buf[0], MMS_HEADER_SIZE );
            if( n < MMS_HEADER_SIZE )
                return -1;
            /* Read metadata */
            if( size > MAX_METADATA_LENGTH )
            {// FIXME reading metadata more then MAX_METADATA_LENGTH bytes
                WEB_MSG( "Metadata size too much: %u\n", size );
                return -1;
            }
            n = read_socket_data( &web_metadata_buf[0], size );
            if( n < size )
                return -1;

            web_metadata_buf[n] = '\0';
            WEB_MSG( "Metadata:\n%s\n", &web_metadata_buf[0] );
            continue;
            break;
        case 'P'/*0x50*/:// $P (Packet-Pair) Packet
        case 'T'/*0x54*/:// $T (Test Data Notification) Packet
            DPRINT( "Packet type: %c\n", web_buf[1] );
            return -1;
            break;
        default:
            web_buf[3] = '\0';
            DPRINT( "Unknown packet type: 0x%02x\n%s\n", web_buf[1], &web_buf[1] );
            return -1;
            break;
        }
        /* Send sound data */
        read_pos = 0;
        do
        {
            block = size - read_pos;
            if( block > BUF_SIZE )
                block = BUF_SIZE;
            n = read_socket_data( &web_buf[0], block );
            if( out_sound_stream( n ) < 0 )
                return -1;
            read_pos += n;
            if( n != block )
                return -1;
        }
        while( read_pos < size );
    }
    return 0;
}
/** Free memory. @param status - exit status. @return status. */
static int free_and_exit(int status)
{
    free_wma_response_parameters();
    return status;
}

int wma()
{
    int r;

    /* Finish work on first response */
    if( com_content_length > 0 && skip_content() < 0 )
        return -1;
    free_http_response_parameters();
    disconnect();
    /* Second request */
    DPRINT("2-d request\n");
    connect_to_server();
    init_wma_response_parameters();
    /* Use parameters from MPlayer */
#define HTTP_REQUEST_STR2 "GET /%s HTTP/1.0\r\n" \
                    "Accept: */*\r\n" \
                    "User-Agent: NSPlayer/4.1.0.3856\r\n" \
                    "Host: %s\r\n" \
                    "Pragma: xClientGUID={c77e7400-738a-11d2-9add-0020af0a3278}\r\n" \
                    "Pragma: no-cache,rate=1.000000,stream-time=0,stream-offset=0:0," \
                            "request-context=1,max-duration=0\r\n" \
                    "Connection: Close\r\n" \
                    "\r\n"
    r = snprintf( web_buf, BUF_SIZE - 1, HTTP_REQUEST_STR2, web_file, web_host );

    DPRINT( web_buf );
    if( send_data( web_buf, r ) < 0 )
        return free_and_exit( -1 );
    /* Reading second response */
    DPRINT("Reading 2-d response, get asf header\n");
    if( read_response( wma_header_line_parser ) < 0 )
        return free_and_exit( -1 );
    if( strstr( web_features, "broadcast") != NULL )
        WEB_MSG("Detected ASF Live stream\n");
    else
        WEB_MSG("Detected ASF Prerecorded\n");

    if
    (
        strnicmp( com_content_type,
            "application/vnd.ms.wms-hdr.asfv1",
            strlen("application/vnd.ms.wms-hdr.asfv1") ) != 0
    )
    {
        // if content type is incorrect, we can also check content.
        // if content not asf header, error
        //if( com_content_length == 0 || check_header( (size_t) r ) < 0 )
        //{
            WEB_MSG( "Incorrect content type:\n%s\n", com_content_type );
            return free_and_exit( -1 );
        //}
    }
    // FIXME change skip_content to check_header
    if( com_content_length > 0 && skip_content() < 0 )
        return free_and_exit( -1 );

    free_wma_response_parameters();
    disconnect();
    /* Third request */
    DPRINT("3-d request\n");
    connect_to_server();
#define HTTP_REQUEST_STR3 "GET /%s HTTP/1.0\r\n" \
                    "Accept: */*\r\n" \
                    "User-Agent: NSPlayer/4.1.0.3856\r\n" \
                    "Host: %s\r\n" \
                    "Pragma: xClientGUID={c77e7400-738a-11d2-9add-0020af0a3278}\r\n" \
                    "Pragma: no-cache,rate=1.000000,stream-time=0,stream-offset=0:0," \
                            "request-context=1,max-duration=0\r\n" \
                    "Pragma: xPlayStrm=1\r\n" \
                    "Pragma: stream-switch-entry=ffff:1:0\r\n" \
                    "Pragma: stream-switch-count=1\r\n" \
                    "Connection: Close\r\n" \
                    "\r\n"

    r = snprintf( web_buf, BUF_SIZE - 1, HTTP_REQUEST_STR3, web_file, web_host );

    DPRINT( web_buf );
    if( send_data( web_buf, r ) < 0 )
        return free_and_exit( -1 );

    /* Reading third response */
    DPRINT("Reading 3-d response, get audio stream\n");
    if( read_response( wma_header_line_parser ) < 0 )
        return free_and_exit( -1 );
    if
    (
        strnicmp( com_content_type,
            "application/x-mms-framed",
            strlen("application/x-mms-framed") ) != 0
    )
    {
        WEB_MSG( "Incorrect content type for audio stream:\n%s\n", com_content_type );
        return free_and_exit( -1 );
    }
    free_wma_response_parameters();

    /* Start playing */
    DPRINT("Start playing\n");
    return play();
}
